from xgboost import XGBClassifier
from sklearn.preprocessing import MinMaxScaler
from sklearn.pipeline import Pipeline
from nyoka import xgboost_to_pmml
from teradataml import create_context
from teradataml.dataframe.dataframe import DataFrame

import joblib
import os


def train(data_conf, model_conf, **kwargs):
    hyperparams = model_conf["hyperParameters"]

    create_context(host=os.environ["AOA_CONN_HOST"],
                   username=os.environ["AOA_CONN_USERNAME"],
                   password=os.environ["AOA_CONN_PASSWORD"])

    feature_names = ["NumTimesPrg", "PlGlcConc", "BloodP", "SkinThick", "TwoHourSerIns", "BMI", "DiPedFunc", "Age"]
    target_name = "HasDiabetes"

    # read training dataset from Teradata and convert to pandas
    train_df = DataFrame(data_conf["table"])
    train_df = train_df.select([feature_names + [target_name]])
    train_df = train_df.to_pandas()

    # split data into X and y
    X_train = train_df.drop(target_name, 1)
    y_train = train_df[target_name]

    print("Starting training...")

    # fit model to training data
    model = Pipeline([('scaler', MinMaxScaler()),
                      ('xgb', XGBClassifier(eta=hyperparams["eta"],
                                            max_depth=hyperparams["max_depth"]))])
    # xgboost saves feature names but lets store on pipeline for easy access later
    model.feature_names = feature_names
    model.target_name = target_name

    model.fit(X_train, y_train)

    print("Finished training")

    # export model artefacts
    joblib.dump(model, "artifacts/output/model.joblib")

    # we can also save as pmml so it can be used for In-Vantage scoring etc.
    xgboost_to_pmml(pipeline=model, col_names=feature_names, target_name=target_name, pmml_f_name="artifacts/output/model.pmml")

    print("Saved trained model")



from __future__ import absolute_import
from typing import Dict

from aoa.api.iterator_base_api import IteratorBaseApi


class DatasetTemplateApi(IteratorBaseApi):
    path = "/api/datasetTemplates/"
    type = "DATASET_TEMPLATE"

    def _get_header_params(self):
        header_vars = ['AOA-Project-ID', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.select_header_accept([
                'application/json',
                'application/hal+json',
                'text/uri-list',
                'application/x-spring-data-compact+json'])]

        return self.generate_params(header_vars, header_vals)

    def find_by_name(self, dataset_template_name: str, projection: str = None):
        """
        returns a dataset template of a project by name

        Parameters:
           dataset_template_name (str): dataset name(string) to find
           projection (str): projection type

        Returns:
            (dict): dataset template
        """
        header_vars = ['AOA-Project-ID', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.select_header_accept([
                'application/json',
                'application/hal+json',
                'text/uri-list',
                'application/x-spring-data-compact+json'])]
        header_params = self.generate_params(header_vars, header_vals)

        query_vars = ['name', 'projection']
        query_vals = [dataset_template_name, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            self.path + "search/findByName",
            header_params,
            query_params)

    def save(self, dataset_template: Dict[str, str]):
        """
        register a dataset template

        Parameters:
           dataset template (dict): dataset template to register

        Returns:
            (dict): dataset template
        """
        header_vars = ['AOA-Project-ID', 'Accept']
        header_vals = [
            self.aoa_client.project_id,
            self.aoa_client.select_header_accept([
                'application/json',
                'application/hal+json',
                'text/uri-list',
                'application/x-spring-data-compact+json'])]
        header_params = self.generate_params(header_vars, header_vals)

        self.required_params(['name', 'templateBody'], dataset_template)

        query_params = {}

        return self.aoa_client.post_request(
            self.path,
            header_params,
            query_params,
            dataset_template)




from __future__ import absolute_import
from typing import List, Dict
from aoa.api_client import AoaClient

import abc


class BaseApi(object):
    path = ''

    def __init__(self, aoa_client=None):
        """
        constructor for api

        Parameters:
           aoa_client (AoaClient): AoaClient to use with api, if not specified will create a new AoaClient object
        """
        if aoa_client is None:
            aoa_client = AoaClient()
        self.aoa_client = aoa_client

    @abc.abstractmethod
    def _get_header_params(self):
        pass

    def generate_params(self, header_params: List[str], header_vals: List[str]):
        """
        returns generated parameters

        Parameters:
           header_params (List[str]): list of parameter names
           header_vals (List[str]): list of parameter values

        Returns:
            (dict): generated parameters
        """
        return dict(zip(header_params, header_vals))

    def required_params(self, param_names: List[str], dict_obj: Dict[str, str]):
        """
        checks required parameters, raises exception if the required parameter is missing in the dictionary

        Parameters:
           param_names (List[str]): list of required parameter names
           dict_obj (Dict[str, str]): dictionary to check for required parameters
        """
        for param in param_names:
            if param not in dict_obj:
                raise ValueError("Missing required value " + str(param))

    def find_all(self, projection: str = None, page: int = None, size: int = None, sort: str = None):
        """
        returns all entities

        Parameters:
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): all entities
        """
        header_params = self._get_header_params()

        query_vars = ['projection', 'sort', 'page', 'size', 'sort']
        query_vals = [projection, sort, page, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            self.path,
            header_params,
            query_params)

    def find_by_archived(self, archived: bool = False, projection: str = None, page: int = None, size: int = None, sort: str = None):
        """
        returns all entities by archived

        Parameters:
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order e.g. name?asc / name?desc
           archived (bool): whether to return archived or unarchived entities

        Returns:
            (dict): all entities
        """
        header_params = self._get_header_params()

        query_vars = ['projection', 'sort', 'page', 'size', 'sort', 'archived']
        query_vals = [projection, sort, page, size, sort, archived]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            self.path + "search/findByArchived",
            header_params,
            query_params)

    def find_by_id(self, id: str, projection: str = None):
        """
        returns the entity

        Parameters:
           id (str): entity id(uuid) to find
           projection (str): projection type

        Returns:
            (dict): entity
        """
        header_params = self._get_header_params()

        query_vars = ['projection']
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.aoa_client.get_request(
            self.path + id,
            header_params,
            query_params)

    def archive(self, id: str):
        """
        archives the entity
        Parameters:
           id (str): entity id(uuid) to archive
           projection (str): projection type
        Returns:
            (dict): entity
        """
        header_params = self._get_header_params()

        return self.aoa_client.post_request(
            "/api/archives/" + self.type + "/" + id,
            header_params,
            {},
            {})

    def unarchive(self, id: str):
        """
        unarchives the entity
        Parameters:
           id (str): entity id(uuid) to unarchive
           projection (str): projection type
        Returns:
            (dict): entity
        """
        header_params = self._get_header_params()

        return self.aoa_client.delete_request(
            "/api/archives/" + self.type + "/" + id,
            header_params,
            {},
            {})



import time
import json
import os
import re

from prometheus_client.metrics import MetricWrapperBase
from prometheus_client import values
from prometheus_client.registry import REGISTRY
from prometheus_client.utils import floatToGoString, INF
from prometheus_client import CollectorRegistry, push_to_gateway

PROMETHEUS_PUSH_GATEWAY_URL = os.getenv('PROMETHEUS_PUSH_GATEWAY_URL', 'localhost:9091')


def publish_data_stats(data_stats, exported_job="aoa_scoring"):
    """
    Publishes the results of record_scoring_stats(..) to prometheus for scoring only.

    This should use the MonitoringHistogram below to record and then push to prometheus the metrics as we do in the
    Java code here https://github.com/ThinkBigAnalytics/AoaCoreService/blob/master/aoa-model-trainer/src/main/java/com/teradata/aoa/trainer/service/monitoring/DataStatisticsPublisher.java#L73

    As we are doing this here in python we may look to remove the same logic from Java and always publish it from here
    for train, eval and scoring.

    Note that we will probably also need to use this for R code later also as R can't publish to prometheus and this will
    allow us to have a single place to publish these metrics.

    :param exported_job:
    :param data_stats:
    :return:
    """

    registry = CollectorRegistry()
    
    # MonitoringHistogram for features and predictors. make sure to take care if categorical or continuous.
    # Just follow logic in Java but it will be much simpler in python :)

    try:
        labels = {
            "model_version": os.environ["MODEL_VERSION"],
            "project_id": os.environ["PROJECT_ID"],
            "model_id": os.environ["MODEL_ID"]
        }
    except Exception as ex:
        raise Exception(f"Missing required environmental variables. Contact system administrator.\nError: {ex}")

    if "features" in data_stats:
        features = data_stats["features"]
        for name, feature in features.items():
            feature_group = feature["group"] if "group" in feature and feature["group"] != "" else "default"
            name = create_metric_name("model_features", name, feature_group)
            if "type" in feature:
                if feature["type"].lower() == "categorical":
                    record_categorical_histogram(registry, name, name, labels, feature)
                elif feature["type"].lower() == "continuous":
                    record_continuous_histogram(registry, name, name, labels, feature)

    if "predictors" in data_stats:
        predictors = data_stats["predictors"]
        for name, predictor in predictors.items():
            predictor_group = predictor["group"] if "group" in predictor and predictor["group"] != "" else "default"
            name = create_metric_name("model_predictors", name, predictor_group)
            if "type" in predictor:
                if predictor["type"].lower() == "categorical":
                    record_categorical_histogram(registry, name, name, labels, predictor)
                elif predictor["type"].lower() == "continuous":
                    record_continuous_histogram(registry, name, name, labels, predictor)

    grouping_key = {
        "model_version": labels['model_version'],
        "job": exported_job
    }
    push_to_gateway(PROMETHEUS_PUSH_GATEWAY_URL, job=exported_job, registry=registry, grouping_key=grouping_key)


def save_evaluation_metrics(metrics):
    with open("artifacts/output/metrics.json", "w+") as f:
        json.dump(metrics, f)


def record_continuous_histogram(registry, documentation, name, labels, data):
    if "statistics" in data and "histogram" in data["statistics"]:
        histogram = data["statistics"]["histogram"]
        edges = histogram["edges"]
        edges.sort()
        value_list = histogram["values"]

        monitor = MonitoringHistogram(name=name,
                                      documentation=documentation,
                                      buckets=edges,
                                      labelnames=["project_id", "model_id", "model_version"],
                                      registry=registry)

        monitor.labels(labels["project_id"],
                       labels["model_id"],
                       labels["model_version"]).record(value_list, bucket_offset=1)


def record_categorical_histogram(registry, documentation, name, labels, data):
    if "statistics" in data and "frequency" in data["statistics"]:
        frequency = data["statistics"]["frequency"]
        categories = [int(x) for x in [*frequency]]
        categories.sort()
        value_list = [*frequency.values()]

        monitor = MonitoringHistogram(name=name,
                                      documentation=documentation,
                                      buckets=categories,
                                      labelnames=["project_id", "model_id", "model_version"],
                                      registry=registry)

        monitor.labels(labels["project_id"],
                       labels["model_id"],
                       labels["model_version"]).record(value_list, bucket_offset=0)


def create_metric_name(metric_prefix, name, group=""):
    return convert_to_valid_name(f"{metric_prefix}_{group}_{name}").lower()


def convert_to_valid_name(name):
    return re.sub('[^a-zA-Z_:][^a-zA-Z0-9_:]*', '', name)


class MonitoringHistogram(MetricWrapperBase):
    """A Custom MonitoringHistogram which we can use to record and publish histograms captured via other systems
    (e.g. Teradata VAL etc) to prometheus for model drift monitoring
    """
    _type = 'histogram'
    _reserved_labelnames = ['le']
    DEFAULT_BUCKETS = (.005, .01, .025, .05, .075, .1, .25, .5, .75, 1.0, 2.5, 5.0, 7.5, 10.0, INF)

    def __init__(self,
                 name,
                 documentation,
                 labelnames=(),
                 namespace='',
                 subsystem='',
                 unit='',
                 registry=REGISTRY,
                 _labelvalues=None,
                 buckets=DEFAULT_BUCKETS
                 ):
        self._prepare_buckets(buckets)
        super(MonitoringHistogram, self).__init__(
            name=name,
            documentation=documentation,
            labelnames=labelnames,
            namespace=namespace,
            subsystem=subsystem,
            unit=unit,
            registry=registry,
            _labelvalues=_labelvalues
        )
        self._kwargs['buckets'] = buckets

    def _prepare_buckets(self, buckets):
        buckets = [float(b) for b in buckets]
        if buckets != sorted(buckets):
            # This is probably an error on the part of the user,
            # so raise rather than sorting for them.
            raise ValueError('Buckets not in sorted order')
        if buckets and buckets[-1] != INF:
            buckets.append(INF)
        if len(buckets) < 2:
            raise ValueError('Must have at least two buckets')
        self._upper_bounds = buckets

    def _metric_init(self):
        self._buckets = []
        self._created = time.time()
        bucket_labelnames = self._labelnames + ('le',)
        self._sum = values.ValueClass(self._type, self._name, self._name + '_sum', self._labelnames, self._labelvalues)
        self._count = values.ValueClass(self._type, self._name, self._name + '_count', self._labelnames, self._labelvalues)
        for b in self._upper_bounds:
            self._buckets.append(values.ValueClass(
                self._type,
                self._name,
                self._name + '_bucket',
                bucket_labelnames,
                self._labelvalues + (floatToGoString(b),))
            )

    def record(self, value_list, bucket_offset):
        """


        categorical: the values from VAL match exactly with the bucket edge as with categorical they are frequency
        and we are abusing histograms here.. However, the way we do it is to place the frequency value for the category
        (ordinal label) into the matching bucket.

        :param value_list: the values for each bucket
        :param bucket_offset: the offset for the value index vs the bucket index (see general method comment)
        :return:
        """
        self._sum.set(0)
        for i, value in enumerate(value_list):

            self._buckets[i+bucket_offset].set(value)
            self._count.inc(value)

    def _child_samples(self):
        samples = []
        acc = 0
        for i, bound in enumerate(self._upper_bounds):
            acc = acc + self._buckets[i].get()
            samples.append(('_bucket', {'le': floatToGoString(bound)}, acc))
        samples.append(('_count', {}, self._count.get()))
        samples.append(('_sum', {}, self._sum.get()))
        samples.append(('_created', {}, self._created))

        return tuple(samples)



from __future__ import absolute_import
from typing import Dict

from aoa.api.iterator_base_api import IteratorBaseApi


class ProjectApi(IteratorBaseApi):

    path = "/api/projects/"
    type = "PROJECT"

    def _get_header_params(self):
        # The header for project id is required for the archive/unarchive method from base_api
        header_vars = ['AOA-Project-ID', 'Accept']
        header_vals = [self.aoa_client.project_id, 'application/json']

        return self.generate_params(header_vars, header_vals)

    def save(self, project: Dict[str, str]):
        """
        create a project

        Parameters:
           project (dict): project to create

        Returns:
            (dict): project
        """
        header_vars = ['Accept']
        header_vals = ['application/json']
        header_params = self.generate_params(header_vars, header_vals)

        query_params = {}

        self.required_params(['description', 'gitRepositoryUrl', 'groupId', 'name'], project)

        return self.aoa_client.post_request(
            self.path,
            header_params,
            query_params,
            project)

    def update(self, project: Dict[str, str]):
        """
        update a project

        Parameters:
           project (dict): project to update

        Returns:
            (dict): project
        """
        header_vars = ['Accept']
        header_vals = ['application/json']
        header_params = self.generate_params(header_vars, header_vals)

        query_params = {}

        self.required_params(['description', 'gitRepositoryUrl', 'groupId', 'name'], project)

        return self.aoa_client.put_request(
            self.path + self.aoa_client.project_id,
            header_params,
            query_params,
            project)

    def is_archived(self, project_id: str):
        """
        returns the archived status of a project
        Parameters:
           project_id (str): project id to check archived status
        Returns:
            (bool)
        """
        project = self.find_by_id(project_id)

        return project["archived"]



from pyspark import SparkConf
from pyspark.sql import SparkSession
from .util import read_dataframe

import logging
import joblib
import pandas as pd
import os


logging.getLogger("py4j").setLevel(logging.ERROR)

# spark name and other properties are set by the framework launcher in spark-submit. don't override
spark = SparkSession.builder \
    .config(conf=SparkConf()) \
    .getOrCreate()


def score(data_conf, model_conf, **kwargs):
    model = joblib.load('artifacts/input/model.joblib')

    test_df = read_dataframe(spark, data_conf["url"])

    # do feature eng in spark / joins whatever reason you're using pyspark...
    # split into test and train
    test_df = test_df.randomSplit([0.7, 0.3], 42)[1].toPandas()

    X = test_df[model.feature_names]

    print("Scoring")
    y_pred = model.predict(X)

    y_pred = pd.DataFrame(y_pred, columns=["pred"])

    # wrap as pyspark df
    predictions = spark.createDataFrame(y_pred)

    # in a real world you would write the results back to HDFS, Teradata, S3 etc.
    predictions.write.mode("overwrite").save("/tmp/predictions")

    logging.info("Finished scoring")


# Add code required for RESTful API
class ModelScorer(object):

    def __init__(self, config=None):
        self.model = joblib.load('artifacts/input/model.joblib')

        from prometheus_client import Counter
        self.pred_class_counter = Counter('model_prediction_classes',
                                          'Model Prediction Classes', ['model', 'version', 'clazz'])

    def predict(self, data):
        pred = self.model.predict([data])

        # record the predicted class so we can check model drift (via class distributions)
        self.pred_class_counter.labels(model=os.environ["MODEL_NAME"],
                                       version=os.environ.get("MODEL_VERSION", "1.0"),
                                       clazz=str(int(pred))).inc()

        return pred



from pyspark import SparkConf
from pyspark.sql import SparkSession
from xgboost import XGBClassifier
from sklearn.preprocessing import MinMaxScaler
from sklearn.pipeline import Pipeline
from nyoka import xgboost_to_pmml
from .util import read_dataframe

import logging
import joblib
import urllib.request

logging.getLogger("py4j").setLevel(logging.ERROR)

# spark name and other properties are set by the framework launcher in spark-submit. don't override
spark = SparkSession.builder \
    .config(conf=SparkConf()) \
    .getOrCreate()


def train(data_conf, model_conf, **kwargs):
    hyperparams = model_conf["hyperParameters"]

    feature_names = ["NumTimesPrg", "PlGlcConc", "BloodP", "SkinThick", "TwoHourSerIns", "BMI", "DiPedFunc", "Age"]
    target_name = "HasDiabetes"

    train_df = read_dataframe(spark, data_conf["url"])

    # do feature eng in spark / joins whatever reason you're using pyspark...
    # split into test and train
    train_df = train_df.randomSplit([0.7, 0.3], 42)[0].toPandas()

    # split data into X and y
    X_train = train_df.drop(target_name, 1)
    y_train = train_df[target_name]

    print("Starting training...")

    # fit model to training data
    model = Pipeline([('scaler', MinMaxScaler()),
                      ('xgb', XGBClassifier(eta=hyperparams["eta"],
                                            max_depth=hyperparams["max_depth"]))])
    # xgboost saves feature names but lets store on pipeline for easy access
    model.feature_names = feature_names

    model.fit(X_train, y_train)

    print("Finished training")

    # export model artefacts
    joblib.dump(model, "artifacts/output/model.joblib")

    # we can also save as pmml so it can be used for In-Vantage scoring etc.
    xgboost_to_pmml(pipeline=model, col_names=feature_names[0:8], target_name=target_name, pmml_f_name="artifacts/output/model.pmml")

    print("Saved trained model")
